
# Utilities
from functools import wraps

class Config():
    ''' Interface shared by all config classes '''

    CONFIG_TYPE = 'config'

    @staticmethod
    def normalize_url(url:str):
        ''' Utility to add a '/' to prefix urls '''

        return '/' + url if url[0] != '/' else url


    def validate_dict(func):
        ''' Decorator used to verify that all arguments passed from_dict() are valid arguments for the class' __init__() method '''
        
        @wraps(func)
        @classmethod
        def wrapped(cls, *args):
            try:
                # Check number of arguments
                assert len(args) == 2, f"{cls.__name__}.from_dict() only passed [{len(args)}] arguments but 2 are required (name and config)"
                # Check types
                config = args[1]; types = cls.__init__.__annotations__
                for key in config: assert isinstance(config[key], types[key]) or not config[key], \
                    f"{cls.__name__}.from_dict() for url [{args[0]}] was passed an invalid type. Key [{key}] must have a type of {types[key]} but is {type(config[key])}"

                return func(cls, args[0], args[1])

            # Catch invalid parameters
            except KeyError as e:
                invalid_param = str(e)[str(e).find("'") + 1: str(e).rfind("'")]
                raise TypeError(f"{cls.__name__}.from_dict() for {cls.CONFIG_TYPE} [{args[0]}] was passed invalid parameter [{invalid_param}]. Supported parameters (config dictionary keys) are {list(filter(lambda x: x!= 'self', cls.__init__.__code__.co_varnames))}")

        return wrapped


    @validate_dict
    def from_dict(cls, name:str, config:dict) -> 'Config':
        ''' Factory for generating a `Config` class from a dictionary. Overriden by concrete config classes '''
        
        return cls(name, **config)


    def to_dict(self):
        ''' Get the route in dictionary form '''

        return self.__dict__


# Interface class
from .setting import Setting

# Utilities
import os, redis

class Redis_Settings(Setting):
    ''' Used to specify and validate Redis settings '''

    CONFIG_TYPE = 'redis_settings'

    # Redis Config
    USE_REDIS = os.environ.get('USE_REDIS', True)
    FORCE_START_REDIS = os.environ.get('FORCE_START_REDIS', True)
    REDIS_HOST = os.environ.get('REDIS_HOST', 'localhost')
    REDIS_PORT = os.environ.get('REDIS_PORT', 6379)
    REDIS_DB = os.environ.get('REDIS_DB', 0)
    REDIS_INSTALLATION_PATH = os.environ.get('REDIS_INSTALLATION_PATH', '/usr/local/bin/redis-server')
    REDIS_CONNECTION_STRING = os.environ.get('REDIS_CONNECTION_STRING')

    def __init__(self, use_redis:bool=None, force_start_redis:bool=None, redis_host:str=None, redis_port:str=None, redis_db:str=None, redis_connection_string:str=None):

        if use_redis: Redis_Settings.USE_REDIS = use_redis
        if force_start_redis: Redis_Settings.FORCE_START_REDIS = force_start_redis
        if redis_host: Redis_Settings.REDIS_HOST = redis_host
        if redis_port: Redis_Settings.REDIS_PORT = int(redis_port)
        if redis_db: Redis_Settings.REDIS_DB = int(redis_db)
        if redis_connection_string:
            Redis_Settings.REDIS_CONNECTION_STRING = redis_connection_string
        else:
            Redis_Settings.REDIS_CONNECTION_STRING = f'redis://{Redis_Settings.REDIS_HOST}:{Redis_Settings.REDIS_PORT}/{Redis_Settings.REDIS_DB}'

        # Allow redis to be force started from the application
        if Redis_Settings.FORCE_START_REDIS and not Redis_Settings.check_redis_connection():
            os.system(f'{Redis_Settings.REDIS_INSTALLATION_PATH} --daemonize yes')

    @staticmethod
    def get_log_data():
        ''' Returns a list of the settings to log to console '''

        redis_online = Redis_Settings.check_redis_connection()
        if Redis_Settings.USE_REDIS: conn_test = 'Connected to Redis :)' if redis_online else 'WARNING - Redis ping failed. Ensure the service is running and config is correct. Defaulting to in-memory cache'
        else: conn_test = 'Redis is disabled via config, defaulting to in-memory cache'
        return [
            f'Redis host set to [{Redis_Settings.REDIS_HOST}]',
            f'Redis port set to [{Redis_Settings.REDIS_PORT}]',
            f'Redis default db set to [{Redis_Settings.REDIS_DB}]',
            conn_test
        ]

    @staticmethod
    def check_redis_connection() -> bool:
        ''' Check if Redis is online '''

        try: # Determine if redis is online
            return redis.Redis(host=Redis_Settings.REDIS_HOST, port=Redis_Settings.REDIS_PORT, db=Redis_Settings.REDIS_DB, socket_connect_timeout=1).ping()
        except:
            return False



# Flask HTTP
from flask import request, Request, Response, jsonify

# API Errors
from .errors import API_Error

# API Utilities
from .utils import *

# App Settings
from dead_simple_framework.config import App_Settings

# Route object
from ..config import Route

# Typing
from typing import Callable, Dict

# TODO - [Useability]    | Authentication request/verification

class API:
    ''' Internal HTTP requests handler for Flask routes '''

    ROUTES:Dict[str, Route] = {}

    @classmethod
    def HANDLER(cls):  # Supported auto-handling for routes
        return {'GET': cls.GET, 'POST': cls.POST, 'PUT': cls.PUT, 'DELETE': cls.DELETE}
    
    @staticmethod
    def GET(request:Request, database:str=None, collection:str=None) -> Response:
        ''' Fetch model data from the server or respond with the appropriate HTTP status code on error. 
        
            The request query string must supply the model's (class) name. It may also supply one or more optional 
            filter and sort parameters.
            --> request : The GET request sent to the server.
            <-- JSON containing the HTTP status code signifying the request's success or failure and
                a list of MongoDB records matching the supplied parameters if the request did not fail.
                GET Request Formats:
                    All - Get all records from a collection.
                        /api/
                        
                    Filtering - Get all records from a collection with field(s) matching the provided value(s).
                        /api/?filter=<<field>>:<<value>>
                        /api/?filter=<<field1>>:<<value1>>,<<field2>>:<<value2>>
                                                                            
                    Sorting - Get all records from a collection. sorted by the provieded field(s). 
                        /api/?sort=<<field>>
                        /api/?sort=<<field1>>,<<field2>>

                    Filtering + Sorting - Get all records from a collection with field(s) matching the provided 
                                        value(s) sorted by the provided field(s).
                        
                        /api/?filter=<<field1>>,<<field2>>:<<value>>&sort=<<field1>> 
                GET Response Format:
                    {"data": [JSON], "code": <<code>>}            
        '''
        
        try:
            # Parse the query string into key/value pairs and store
            payload = parse_query_string(request.query_string.decode()) if request.query_string.decode() else {}
            # Use the query string to send a database query
            data_cursor = fetch_and_filter_data(payload, database, collection)
            # Sort the data if one was specified in the query string
            sorted_data = list(sort_data(data_cursor, payload))

            return JsonResponse({'data': sorted_data}, 200 if len(sorted_data) > 0 else 404)
            
        except Exception as e:
            if App_Settings.APP_ENV == 'development':
                raise e

            return JsonError('GET', e)


    @staticmethod
    def POST(request:Request, database:str=None, collection:str=None) -> Response:
        ''' Create a new MongoDB record or respond with the appropriate HTTP status code on error. 
            
            --> request : The POST request sent to the server.
            <-- JSON containing the HTTP status code signifying the request's success or failure.
                POST Body Format:
                        { {<<field1>>: <<value1>>, <<field2>>: <<value2>>} }
                POST Response Format:
                    {"code": <<code>>}            
        '''

        try:
            # Get request body parameters
            payload = request.get_json(force=True) if request.data else {}
            # Remove any passed _id and insert in the database [TODO - allow? Maybe a config option?]
            if payload:
                payload.pop('_id', None)
                inserted_id = insert_data(payload, database, collection)
            else:
                raise API_Error('No data supplied to POST', 500)

            return JsonResponse({'_id': str(inserted_id)}, code=200)

        except Exception as e:
            if App_Settings.APP_ENV == 'development':
                raise e

            return JsonError('POST', e)


    @staticmethod
    def PUT(request:Request, database:str=None, collection:str=None) -> Response:
        ''' Update a MongoDB record or respond with the appropriate HTTP status code on error. 
            
            The request body must supply a dictionary of 
            one or more field/value pairs to update the record with, along with the ID of the record to update
            
            --> request : The PUT request sent to the server.
            <-- JSON containing the HTTP status code signifying the request's success or failure.
                PUT Body Format:
                    { 
                        "filter": {<<field1>>: <<value1>>, <<field2>>: <<value2>>},
                        "fields": {<<field1>>: <<value1>>, <<field3>>: <<value3>>} 
                    }
                PUT Response Format:
                    {"code": <<code>>}            
        '''

        try:
            # Get request body parameters
            payload = request.get_json(force=True) if request.data else {}
            # Use the passed _id to update data in the database
            if payload:
                _id = payload.get('_id')
                if not getattr(update_data(payload, database, collection), 'modified_count', None): 
                    raise API_Error(f'ID [{_id}] not found', 404)
            else:
                raise API_Error('No data supplied to PUT', 500)

            return JsonResponse(code=200)

        except Exception as e:
            if App_Settings.APP_ENV == 'development':
                raise e

            return JsonError('PUT', e)


    @staticmethod
    def DELETE(request:Request, database:str=None, collection:str=None) -> Response:
        ''' Delete a MongoDB record or respond with the appropriate HTTP status code on error. 
            
            The request body must supply the ID of the record to delete
            *The filter parameters must match only a single model*
            
            --> request : The DELETE request sent to the server.
            <-- JSON containing the HTTP status code signifying the request's success or failure.
                DELETE Body Format:
                    { "model": <<model>>, "filter": {<<field1>>: <<value1>>, <<field2>>: <<value2>>} }
                DELETE Response Format:
                    {"code": <<code>>}            
        '''

        try:
            # Get request body parameters
            payload = request.get_json(force=True) if request.data else {}
            if payload:
                # Use the passed _id to update data in the database
                _id = payload.get('_id')
                if not getattr(delete_data(payload, database, collection), 'deleted_count', None):
                    raise API_Error(f'ID [{_id}] not found', 404)
            else:
                raise API_Error('No data supplied to DELETE', 500)

            return JsonResponse(code=200)

        except Exception as e:
            if App_Settings.APP_ENV == 'development':
                raise e

            return JsonError('DELETE', e)


    @classmethod
    def _check_method(cls, method:str, route:Route):
        ''' Ensure the method is allowed for a route or raise a 405 error '''

        # If the method isn't listed in the route conig the user passed it isn't allowed
        if method not in route.methods:
            raise API_Error(f'Method [{method}] not allowed for route [{route.url}]', 405)

        # If the method doesn't have a handler assigned it isn't allowed
        if method not in cls.HANDLER():
            raise API_Error(f'No hanlder for [{method}]', 405)

    
    @classmethod
    def _check_logic(cls, route_name:str, logic_func:Callable, collection:str):
        ''' Ensure user defined logic '''

        num_parameters = len(logic_func.__code__.co_varnames)

        if collection and num_parameters < 3:
            raise API_Error(f'Handler [{logic_func.__name__}] for route [{route_name}] supports [{num_parameters}] arguments. Must support 3 (request, payload, collection)', 500)
        elif num_parameters < 2:
            raise API_Error(f'Handler [{logic_func.__name__}] for route [{route_name}] supports [{num_parameters}] arguments. Must support 2 (request, payload)', 500)
        

    @classmethod
    def main(cls) -> Response:
        ''' Called when the /api/ endpoint is sent an HTTP request. Delegates 
            to the appropriate handler based on the request method or returns a JSON
            formatted error if the method is not supported.
            --> request : The HTTP request sent to the server.
            <-- JSON containing the HTTP status code signifying the request's success or failure and
                all other data returned from the server.
        '''

        # Get the configuration for the route that was accessed
        route = cls.ROUTES[str(request.url_rule)]

        # Ensure the method is allowed
        cls._check_method(request.method, route)

        # Get the collection
        database, collection = route.database, route.collection

        # Get the logic that should be run for this route (if any)
        logic = route.logic
        
        if not logic: # Use default handling if no logic provided
            # Run CRUD handling
            return cls.HANDLER()[request.method](request, database=database, collection=collection)

        if str(request.method) == 'GET':
            payload = parse_query_string(request.query_string.decode()) if request.query_string.decode() else {}
        else:
            payload = request.get_json(force=True) if request.data else dict(request.form)

        # Ensure user defined logic can accept arguments to pass or throw a warning
        cls._check_logic(route.name, logic, collection)

        # If a collection is specified, pass through to next function
        if route.collection or route.database:
            with Database(database=route.database, collection=route.collection) as collection:
                return logic(request, payload, collection)
        
        # Otherwise just pass the request
        return logic(request, payload)



from .route import Route
from .task import TaskConfig
from .settings.app_settings import App_Settings
from .settings.rabbitmq_settings import RabbitMQ_Settings
from .settings.redis_settings import Redis_Settings
from .settings.mongodb_settings import MongoDB_Settings


''' Shared utility functions for task result storage/caching '''

# Database
from ..database import Database
from bson import ObjectId

# Cache
from ..cache import Cache


def cache_mongo_id(_id:ObjectId, cache_key:str, cache_sub_key:str=None):
    ''' Cache the ID of persistently stored data '''

    Cache().cache_dynamic_dict(cache_key, {cache_sub_key: str(_id)}) if cache_sub_key else Cache().cache_string(cache_key, str(_id))


def insert_persistently_and_cache(collection:str, cache_key:str, document:dict, cache_sub_key=None):
    ''' Store data persistently in the database and make it easily available via the cache '''

    with Database(collection=collection) as db:
        cache_mongo_id(db.insert_one(document).inserted_id, cache_key, cache_sub_key)


def upsert_persistently_and_cache(collection:str, cache_key:str, document:dict, cache_sub_key=None):
    ''' Update data stored persistently or insert if no data stored '''

    cache = Cache()
    if cache_sub_key:
        _id = cache.get_dynamic_dict_value(cache_key, cache_sub_key) or ObjectId()
    else:
        _id = cache.get(cache_key) or ObjectId()

    with Database(collection=collection) as db:
        db.delete_one({'_id': ObjectId(_id)})
        cache_mongo_id(db.insert_one(document).inserted_id, cache_key, cache_sub_key)



# Flask
from flask import Blueprint

# Internal API 
from .api import API

# Route config class
from .config import Route

# Encoding
import json

# Typing
from typing import Union

# Debug
from pprint import pprint

# TODO - [Useability]     | RouteConfig class to allow a route to be created with hinting/linting enabled
# TODO - [Extendability] | Allow `logic` as default and method specific logic

class Router:
    ''' Module allowing route specification as a list of dictionaries in config/routes.py'''

    @staticmethod
    def _cast_route(url:str, route_config:Union[dict, Route]) -> Route:
        ''' Takes a route config dictionary and casts it to a `Route` object 
        
        Args:
            url (str): The url for this route
            route_config (Union[dict, Route]): The configuration for the route. Can be passed a config dictionary
                or an already created route. If the latter if passed, no conversion is necessary
        Returns:
            A `Route` object representing the config for that route
        '''
        
        return route_config if isinstance(route_config, Route) else Route.from_dict(url, route_config)


    @classmethod
    def register_routes(cls, app, routes:dict):   
        ''' Register all routes in config/routes.py '''
 
        route_objs ={}
        for route_url, route in routes.items():
            # Cast config to `Route` object if it's a dictionary
            route = cls._cast_route(route_url, route)
            # Create a Flask blueprint based on the config for this route in the `routes` dictionary
            blueprint = Blueprint(route.name, __name__)
            # Then attach logic and URL, and other options specified in the config to the blueprint
            cls.configure_blueprint(route, blueprint)
            # Register the blueprint with the Flask app so HTTP requests can be directed to it's URL
            app.register_blueprint(blueprint)
            # Add the Route object to the internal registry
            route_objs[route.url] = route

        API.ROUTES = route_objs # Copy routes to the internal API cass it can reference them

    @staticmethod
    def configure_blueprint(route:Route, blueprint: Blueprint):
        ''' Adds configurations to the route Blueprint based on the dictionary specification '''

        # Ensure logic or a collection is set
        assert route.collection or route.logic, f"No collection or logic specified for route {route.url}. This route will do nothing!"

        # Set the blueprint to the URL specified in the route configuration
        blueprint.add_url_rule(route.url, route.name, view_func=API.main, methods=route.methods, **({'defaults': route.defaults} if route.defaults else {}))



# Database
from ..database import Database

# Celery
from celery import Task

# Utilities
from .utils import insert_persistently_and_cache, upsert_persistently_and_cache

# Errors
from pymongo.errors import InvalidDocument

# Debug
import logging

# TODO - [Logging]       | Combine failure logging with main app stderr if possible?
# TODO - [Stability]     | Automatic retry
# TODO - [Useability]    | Storage of failed tasks for easy retreival [configurable]
# TODO - [Extendability] | Allow results collection and cache key to be overridden
# TODO - [Extendability] | Allow results to be cached directly [new task class]


class Database_Task(Task):
    ''' Base Celery task for storing results in the database '''

    _collection  = '_task_results_'  # Collection to store results in
    _cache_key =   '_task_results_'  # Cache key to store latest result ID (so you don't need to watch Mongo collections)
    _is_setup = False

    # TODO - Distinct task ID field?

    def _setup(self):
        ''' Perform database optimizations (once) '''

        # Add index for known lookups
        with Database(collection=self._collection) as db:
            db.create_index([('task_name', 1), ('task_id', -1)])

        self._is_setup = True


    def on_success(self, retval, task_id, args, kwargs):
        ''' Stores the result of an asynchronous task in Mongo when it completes '''

        if not self._is_setup: 
            self._setup()                               # Perform initial setup if not done
        
        self.success(retval, task_id, args, kwargs)     # Delegate storage logic to concrete subclasses
    
    
    def success(self, retval, task_id, args, kwargs):
        ''' Abstract - Implemented by concrete Task classes '''


class Store_Task(Database_Task):
    ''' Celery task that persistently stores the all results in the database '''

    def success(self, retval, task_id, args, kwargs):
        ''' Stores the result of an asynchronous task in Mongo when it completes '''

        data = {'task_name': self.name, 'task_id': str(task_id), 'task_result': retval}
        try:
            # Store the result in MongoDB for retrieval with `Task_Manager.get_result()`
            insert_persistently_and_cache(self._collection, self._cache_key, data, self.name)
        except InvalidDocument as e:
            # Coerce task result to string if it can't be serialized 
            data = {'task_name': self.name, 'task_id': str(task_id), 'task_result': str(retval)}
            insert_persistently_and_cache(self._collection, self._cache_key, data, self.name)
        

class Store_Latest_Task(Database_Task):
    ''' Celery task that persistently stores the latest result in the database '''

    def success(self, retval, task_id, args, kwargs):
        ''' Stores the result of an asynchronous task in Mongo when it completes '''

        data = {'task_name': self.name, 'task_id': str(task_id), 'task_result': retval}
        try:
            # Store the result in MongoDB for retrieval with `Task_Manager.get_result()`
            upsert_persistently_and_cache(self._collection, self._cache_key, data, self.name)
        except InvalidDocument as e:
            # Coerce task result to string if it can't be serialized 
            data = {'task_name': self.name, 'task_id': str(task_id), 'task_result': str(retval)}
            upsert_persistently_and_cache(self._collection, self._cache_key, data, self.name)
        



# Interface class
from .config import Config

# Typing
from typing import Callable


class Route(Config):
    ''' Used to specify a route in the config dictionary '''

    SUPPORTED_HTTP_METHODS = ['GET', 'POST', 'DELETE', 'PUT', 'PATCH', 'OPTIONS']
    CONFIG_TYPE = 'url'

    def __init__(self, url:str, name:str=None, methods:list=['GET'], defaults:dict=None, logic:Callable=None, collection:str=None, database:str=None):
        ''' Initialize a new route to add to the route config 
        
        Args:
        
            url (str): The url for this route

            methods (list): The supported methods for this route (must be included in `SUPPORTED_HTTP_METHODS`).
                Defaults to ['GET']

            defaults (dict, optional): Optional defaults for request body or query string parameters for this route
            
            logic (Callable, optional): Optional function to run when a request is made to this route.
                receives the query string or body params with any defaults if they were passed
            
            collection (str, optional): Optional MongoDB collection to use for automatic storage and retrieval
                of data when a request is sent to the route. This allows automatic CRUD operations with no additional
                config. See the documentation for the internal API for more info (TODO)
            
            database (str, optional): Optional MongoDB database to use for automatic storage and retrieval
                of data when a request is sent to the route. This allows automatic CRUD operations with no additional
                config. If this is not set the default database is used. See the documentation for the internal API for more info (TODO)
        '''

        self.url = Config.normalize_url(url)
        self.name = self.url if not name else name
        self.methods = self._validate_methods(methods)
        self.defaults = defaults
        self.logic = logic
        self.collection = collection
        self.database = database


    def _validate_methods(self, methods:list):
        ''' Ensure passed methods are part of `SUPPORTED_HTTP_METHODS` '''

        unsupported_methods = list(set(methods) - set(self.SUPPORTED_HTTP_METHODS))
        assert not unsupported_methods, f"{self.__class__.__name__} was initialized with unsupported `methods` argument. HTTP methods {unsupported_methods} are not supported"        
        return methods



# Flask HTTP
from flask import jsonify, Response

# API Errors
from .errors import API_Error

# Database
from ..database import Database
from pymongo.cursor import Cursor
from bson import ObjectId

# Encoding
import json

# Debug
import logging


def JsonResponse(content: dict = {}, code: int = 200) -> Response:
    ''' Format an API JSON response.
        --> content [dict] : A dictionary of JSON serializable objects to return. Optional.
        
        --> code [int]: The HTTP status code to return. Optional.
        
        <-- The JSON formatted response. { <<content>>, "code": <<code>> }
    '''

    content['code'] = code
    
    return jsonify(**content)


def JsonError(method: str, exception: Exception) -> Response:
    ''' Format an API JSON error response.
        --> method : The HTTP request method that generated the error (e.g. POST).
        
        --> exception : The error generated.
        
        <-- A JSON formatted error response : { "msg" <<error_message>>, "code": <<code>> }.
    '''

    error_msg = '{} - {}'.format(method, str(exception)) 
    error_code = exception.code if type(exception) == API_Error else 500
    
    logging.critical(error_msg)

    return JsonResponse({'msg': error_msg}, error_code)


def parse_query_pairs(payload: str) -> dict:
    ''' Parse out one or more key/value pairs from a query string
        (e.g. /api/?param=key:value || /api/?param=key1:value1,key2:value2).
    
        --> payload : The arguments supplied with the parameter (The key value pairs in string form).
        <-- The parsed key/value pairs.
    '''

    pairs = {}
    for pair_tuple in map(lambda pair: pair.split(':'), [pair for pair in payload.split(',')]):
        pairs[pair_tuple[0]] = pair_tuple[1]

    return pairs


def parse_query_params(payload: str) -> list:
    ''' Parse out one or more parameter values pairs from a query string
        (e.g. /api/?param=value || /api/?param=value1,value2).
    
        --> payload : The arguments supplied with the parameter (The values in string form).
        <-- The parsed values.
    '''

    return [(value, 1) for value in payload.split(',')]


def parse_query_string(payload: str) -> dict:
    ''' Parse out the filter and sort from a query string
    
        --> payload : The arguments supplied with the parameter (The values in string form).
        <-- The parsed values.
    '''

    dict_payload = {}
    query_params = payload.split('&')
    for param in query_params:
        args = param.split('=')
        if args[0] == 'filter':
            dict_payload[args[0]] = parse_query_pairs(args[1])
        elif args[0] == 'sort':
            dict_payload[args[0]] = parse_query_params(args[1])
        else:
            dict_payload[args[0]] = args[1]
        
    return dict_payload


def create_query_string(query_params:dict) -> str:
    ''' Generate a query string from a dictionary of parameters 

        e.g. {'item': '56', 'type': '21' }  -> "?item=56&type=21"
    '''

    return '?' + '&'.join([f"{k}={v}" for k,v in query_params.items()]) if query_params else ''


def fetch_and_filter_data(request_params: dict, database:str=None, collection:str=None) -> list:
    ''' Fetch records from the database matching a filter supplied in an HTTP request.
        Ensure fields supplied in the filter exist for the model. If no filter is supplied
        all objects are retrived.
    
        --> request_params : The parameters sent with the request (in querystring or body).
        <-- A list containing the MongoDB data matching the supplied filter or all objects in a collection.
    '''

    mongo_filter = request_params.get('filter', {}) 
    if request_params.get('sort'):
        [mongo_filter.update({s[0]: {'$exists': True}}) for s in request_params.get('sort')]

    with Database(database,collection) as db:
        return db.find(mongo_filter)
    

def sort_data(data: Cursor, request_params: dict) -> list:
    ''' Sorts a data according to the parameters sent in an HTTP request.
        --> data : The cursor of data to apply the sort to.
        --> request_params : The parameters sent with the request (in querystring or body).
        <-- A queryset containing the sorted data.
    '''

    mongo_sort = request_params.get('sort', {})
    if mongo_sort != {}:
        return data.sort(mongo_sort)
    
    return data


def insert_data(request_params: dict, database:str=None, collection:str=None) -> str:
    ''' Add data to the collection with the parameters sent in an HTTP request.
        --> request_params [dict] : The parameters sent with the request (in querystring or body).
        <-- [str] The ObjectId of the inserted data
    '''

    mongo_fields = request_params.copy()
    with Database(database,collection) as db:
        return str(db.insert_one(mongo_fields).inserted_id)


def update_data(request_params: dict, database:str=None, collection:str=None) -> bool:
    ''' Add data to the collection with the parameters sent in an HTTP request.
        --> request_params [dict] : The parameters sent with the request (in querystring or body).
        <-- [bool] True if successful
    '''

    mongo_fields = request_params.copy()
    _id = mongo_fields.pop('_id', None)
    
    if _id:
        with Database(database,collection) as db:
            return db.update_one({'_id': ObjectId(_id)}, {'$set': mongo_fields}).acknowledged
    else:
        raise API_Error('No ID supplied', 400)


def delete_data(request_params: dict, database:str=None, collection:str=None) -> bool:
    ''' Delete data from the collection with the parameters sent in an HTTP request.
        --> request_params [dict] : The parameters sent with the request (in querystring or body).
        <-- True if records were deleted
    '''

    mongo_fields = request_params.copy()
    if mongo_fields.get('_id'):
        with Database(database,collection) as db:
            return db.delete_one({'_id': ObjectId(mongo_fields.pop('_id', None))}).deleted_count > 0
    else:
        raise API_Error('No ID supplied', 400)



# Interface class
from .setting import Setting

# Utilities
import os, pymongo

class MongoDB_Settings(Setting):
    ''' Used to specify and validate MongoDB settings '''

    CONFIG_TYPE = 'mongodb_settings'

    # MongoDB Config
    MONGODB_HOST = os.environ.get('MONGODB_HOST', 'localhost')
    MONGODB_PORT = os.environ.get('MONGODB_PORT', 27017)
    MONGODB_USERNAME = os.environ.get('MONGODB_USERNAME', '')
    MONGODB_PASSWORD = os.environ.get('MONGODB_PASSWORD', '')
    MONGODB_DEFAULT_DB = os.environ.get('MONGODB_DEFAULT_DB', 'db')
    MONGODB_DEFAULT_COLLECTION = os.environ.get('MONGODB_DEFAULT_COLLECTION', 'data')
    MONGODB_CONNECTION_STRING = os.environ.get('MONGODB_CONNECTION_STRING')
    MONGODB_DATA_PATH = os.environ.get('MONGODB_DATA_PATH', os.path.expanduser('~/data/db'))
    MONGODB_LOG_PATH = os.environ.get('MONGODB_LOG_PATH', os.path.expanduser('/tmp/mongo.log'))
    MONGODB_INSTALLATION_PATH = os.environ.get('RABBITMQ_INSTALLATION_PATH', '/usr/local/bin/mongod')
    FORCE_START_MONGODB = os.environ.get('FORCE_START_MONGODB', True)

    def __init__(self, mongodb_host:str=None, mongodb_port:str=None, mongodb_username:str=None,
                    mongodb_password:str=None, mongodb_default_db:str=None, mongodb_default_collection:str=None,
                    mongodb_connection_string:str=None, mongodb_data_path:str=None, mongodb_log_path:str=None, 
                    force_start_mongodb:bool=None, mongodb_installation_path:str=None):

        if mongodb_host: MongoDB_Settings.MONGODB_HOST = mongodb_host
        if mongodb_port: MongoDB_Settings.MONGODB_PORT = mongodb_port
        if mongodb_username: MongoDB_Settings.MONGODB_USERNAME = mongodb_username
        if mongodb_password: MongoDB_Settings.MONGODB_PASSWORD = mongodb_password
        if mongodb_default_db: MongoDB_Settings.MONGODB_DEFAULT_DB = mongodb_default_db
        if mongodb_default_collection: MongoDB_Settings.MONGODB_DEFAULT_COLLECTION = mongodb_default_collection
        if mongodb_data_path: MongoDB_Settings.MONGODB_DATA_PATH = mongodb_data_path
        if mongodb_log_path: MongoDB_Settings.MONGODB_LOG_PATH = mongodb_log_path
        if mongodb_installation_path: MongoDB_Settings.MONGODB_INSTALLATION_PATH = mongodb_installation_path
        if force_start_mongodb: MongoDB_Settings.FORCE_START_MONGODB = force_start_mongodb
        if mongodb_connection_string: 
            MongoDB_Settings.MONGODB_CONNECTION_STRING = mongodb_connection_string
        else:
            auth = f"{MongoDB_Settings.MONGODB_USERNAME}:{MongoDB_Settings.MONGODB_PASSWORD}@"
            MongoDB_Settings.MONGODB_CONNECTION_STRING = f'mongodb://' + (auth + MongoDB_Settings.MONGODB_HOST if auth.strip() != ':@' else MongoDB_Settings.MONGODB_HOST)

        # Allow MongoDB to be force started
        MongoDB_Settings.check_mongodb_connection(MongoDB_Settings.FORCE_START_MONGODB)
            
    @staticmethod
    def get_log_data():
        ''' Returns a list of the settings to log to console '''

        MongoDB_Settings.check_mongodb_connection()
        return [
            f'MongoDB connection string set to [{MongoDB_Settings.MONGODB_CONNECTION_STRING}:{MongoDB_Settings.MONGODB_PORT}]',
            f'MongoDB default database set to [{MongoDB_Settings.MONGODB_DEFAULT_DB}]',
            f'MongoDB default collection set to [{MongoDB_Settings.MONGODB_DEFAULT_COLLECTION}]',
            f'MongoDB data path set to [{MongoDB_Settings.MONGODB_DATA_PATH}]',
            f'MongoDB log path set to [{MongoDB_Settings.MONGODB_LOG_PATH}]',
            'Connected to MongoDB :)'
        ]

    @staticmethod
    def check_mongodb_connection(can_force_start:bool=False) -> bool:
        ''' Check if MongoDB is online '''

        try: # Determine if MongoDB is online
            return True if pymongo.MongoClient(host=MongoDB_Settings.MONGODB_CONNECTION_STRING, port=MongoDB_Settings.MONGODB_PORT, serverSelectionTimeoutMS=1).server_info() else False
        except:
            if not can_force_start:
                raise Exception(f'ERROR - MongoDB ping failed for host [{MongoDB_Settings.MONGODB_CONNECTION_STRING}:{MongoDB_Settings.MONGODB_PORT}]. Ensure the service is running and config is correct')
            
            print(f'ERROR - MongoDB ping failed for host [{MongoDB_Settings.MONGODB_CONNECTION_STRING}:{MongoDB_Settings.MONGODB_PORT}]. Attempting to force start...')

            if not os.path.exists(f'{MongoDB_Settings.MONGODB_LOG_PATH}'): os.system(f'touch {MongoDB_Settings.MONGODB_LOG_PATH}')
            os.system(f'{MongoDB_Settings.MONGODB_INSTALLATION_PATH} --fork --logpath={MongoDB_Settings.MONGODB_LOG_PATH} --dbpath={MongoDB_Settings.MONGODB_DATA_PATH}')
            MongoDB_Settings.check_mongodb_connection()

            print('Done :)')



# Interface class
from .config import Config

# Typing
from typing import Callable


class TaskConfig(Config):
    ''' Used to specify a task in the config dictionary '''

    CONFIG_TYPE = 'task'

    def __init__(self, name:str, logic:Callable, schedule:dict=None, default_args:tuple=None, default_kwargs:dict=None, depends_on:str=None, store_results:bool=None):
        ''' Initialize a new task to add to the task config 
        
        Args:
            
            name (str): The name for this task

            logic (Callable): The function to run when this task is executed

            schedule (dict): If passed allows the task to run on a schedule (see the `Schedule` config class for more info)
            
            default_args (tuple): A tuple of default arguments to pass to the task
            
            default_kwargs (dict): A dictionary of default keyword arguments + values to pass to the task

            depends_on (str): The name of another task that this one depends on. Used to create chains of dependant tasks

            store_results (bool): If set to True, all results for this task will be store. If set to false no results will be stored.
                If not set, only the latest result will be saved (default behavior)
        '''

        self.name = name
        self.logic = logic
        self.schedule = schedule # TODO - Add schedule config
        self.default_args = default_args
        self.default_kwargs = default_kwargs
        self.depends_on = depends_on
        self.store_results = store_results

        # Set externally
        self.task = None

        # Set the name of the function for the task to the task's name
        self.logic.__name__ = name

    def set_task(self, task):
        ''' Set the actual Celery task for this class '''

        assert not self.task, f"Task.set_task() error for task [{self.name}] | A Celery task is already set!"

        self.task = task



# Requests
import requests

# API Utilities
from .utils import create_query_string

# App Settings
from dead_simple_framework.config import App_Settings

# API Errors
from .errors import API_Client_Error

# Utilities
from time import sleep

# TODO - [Stability]     | Timeouts

class API:
    ''' Client for making HTTP requests  '''

    @classmethod
    def send_request(cls, method:str, url:str, query_params:dict=None, data:dict=None, ignore_errors=False, retry_ms=500, num_retries=3):
        ''' Shared logic that handles actually firing a request '''

        # Format the URL
        url = f"{url}{create_query_string(query_params)}".strip()
        result = None
        
        # Open a session
        with requests.Session() as session:
            # Try the request up to `num_retries` times
            for _ in range(0, num_retries):
                result = session.send(session.prepare_request(requests.Request(method, url, headers=App_Settings.APP_API_CLIENT_HEADERS)))
                if result.status_code == 400: # Fetch again on 400's
                    result = session.send(session.prepare_request(requests.Request(method, url, headers=App_Settings.APP_API_CLIENT_HEADERS)))

                # If there is a result
                if result:
                    # If the status code is not 200, the result may be incorrect
                    if result.status_code != 200:
                        if not ignore_errors:
                            raise API_Client_Error(url, method, result.status_code)

                    # Check for actual result data and return if present
                    if result.text:
                        return result

                # Otherwise wait, then rety
                sleep(retry_ms/1000)
        
        return result if result and result.text else None


    @classmethod
    def options(cls, url:str, query_params:dict=None, data:dict=None, ignore_errors=False, retry_ms=500, num_retries=3):
        ''' [HTTP OPTIONS] Send an options request '''

        return cls.send_request('OPTIONS', url, query_params=query_params, data=data, ignore_errors=ignore_errors, retry_ms=retry_ms, num_retries=num_retries)


    @classmethod
    def post(cls, url:str, query_params:dict=None, data:dict=None, ignore_errors=False, retry_ms=500, num_retries=3):
        ''' [HTTP POST] Send data to an API'''

        return cls.send_request('POST', url, query_params=query_params, data=data, ignore_errors=ignore_errors, retry_ms=retry_ms, num_retries=num_retries)


    @classmethod
    def put(cls, url:str, query_params:dict=None, data:dict=None, ignore_errors=False, retry_ms=500, num_retries=3):
        ''' [HTTP PUT] Replace data via API'''

        return cls.send_request('PUT', url, query_params=query_params, data=data, ignore_errors=ignore_errors, retry_ms=retry_ms, num_retries=num_retries)


    @classmethod
    def patch(cls, url:str, query_params:dict=None, data:dict=None, ignore_errors=False, retry_ms=500, num_retries=3):
        ''' [HTTP PATCH] Update data via API'''

        return cls.send_request('PATCH', url, query_params=query_params, data=data, ignore_errors=ignore_errors, retry_ms=retry_ms, num_retries=num_retries)


    @classmethod
    def get(cls, url:str, query_params:dict=None, ignore_errors=False, retry_ms=500, num_retries=3):
        ''' [HTTP GET] Fetch a resource from an API '''

        return cls.send_request('GET', url, query_params=query_params, ignore_errors=ignore_errors, retry_ms=retry_ms, num_retries=num_retries)


    @classmethod
    def get_json(cls, url:str, query_params:dict=None, ignore_errors=False, retry_ms=500, num_retries=3) -> dict:
        ''' [HTTP GET] Fetch a JSON from an API '''

        # Get the raw result
        res = cls.get(url, query_params, ignore_errors, retry_ms, num_retries)

        # Extract data into dictionary
        return res.json() if res else {}



# Interface class
from .setting import Setting

# Utilities
import os

class App_Settings(Setting):
    ''' Used to specify and validate application settings '''

    CONFIG_TYPE = 'app_settings'

    # App Config
    APP_ENV = os.environ.get('APP_ENV', 'development')
    APP_ENABLE_CORS = os.environ.get('APP_ENABLE_CORS', False)
    APP_HOST = os.environ.get('APP_HOST', '0.0.0.0')
    APP_API_CLIENT_HEADERS = {"User-Agent": "Mozilla/5.0"}
    APP_LOG_CONFIG = os.environ.get('APP_LOG_CONFIG', True)
    APP_DEBUG_MODE = True

    def __init__(self, app_env:str=None, app_enable_cors:bool=None, app_host:str=None, app_api_client_headers:dict=None, app_log_config:bool=None):
        if app_env: App_Settings.APP_ENV = app_env
        os.environ['FLASK_ENV'] = App_Settings.APP_ENV

        if app_enable_cors != None: App_Settings.APP_ENABLE_CORS = app_enable_cors
        if app_host: App_Settings.APP_HOST = app_host
        if app_api_client_headers: App_Settings.APP_API_CLIENT_HEADERS = app_api_client_headers
        if app_log_config: App_Settings.APP_LOG_CONFIG = app_log_config
        
        App_Settings.APP_DEBUG_MODE = True if App_Settings.APP_ENV in ['dev', 'development', 'uat'] else False
        os.environ['FLASK_DEBUG'] = '1' if App_Settings.APP_DEBUG_MODE else '0'


    @staticmethod
    def get_log_data():
        ''' Returns a list of the settings to log to console '''

        return [
            'CORS enabled for application' if App_Settings.APP_ENABLE_CORS else 'CORS disabled for application',
            f'Default API client headers are {App_Settings.APP_API_CLIENT_HEADERS}'
        ]


# Flask
from flask import Flask 
from flask_cors import CORS

# Routing
from .router import Router

#Encoding
from .encoder import JSON_Encoder

# Async Tasks
from .tasks import Task_Manager

# App-wide settings
from .config.settings.main import Settings

# Utilities
import os

# Debug
import logging

# TODO - [Logging] | Configure application wide logging
# TODO - [Useability] | Use `Task_Manager` via adapter to hide unwanted methods/attributes 

class Application(Task_Manager):
    ''' Main application driver

        Takes a configuration dictionary that can specify:
          - HTTP routes             [URL, Allowed Methods, Custom Logic]
          - Asynchronous Tasks      [Schedule, Cacheable, Default Args/Kwargs]

        Runs a server based on that configuration
     '''

    _app = None     # Internal reference - Hacky way to allow classmethods to access a global state (probably a bad idea)

    def __init__(self, config:dict):
        ''' Initialize the server based on the configuration dictionary '''

        Settings(**config.get('settings', {})) 

        # Create Flask application
        self.app = Flask(__name__)

        # Log 3rd party configuration
        Settings.log_config()

        # Register a custom encoder for JSON serialization
        # [Note] - This currently just bypasses errors by casting the offending data to a string :)
        self.app.json_encoder = JSON_Encoder

        # Register all HTTP routes that should be available based on the application config
        Router.register_routes(self.app, config['routes'])
        
        # Initialize inherited asynchronous task management ability
        super().__init__(dynamic_tasks=config.get('tasks'))

        if Settings.APP_ENABLE_CORS: CORS(self.app)

        Application._app = self


    def run(self):
        ''' Runs the server '''

        #self.app.run(host=os.environ.get('FLASK_RUN_HOST', '0.0.0.0'), debug=(os.environ.get('APP_DEBUG') == 'True'))
        self.app.run(
            host=Settings.APP_HOST, 
            debug=Settings.APP_DEBUG_MODE
        )



# Base Encoder
from flask.json import JSONEncoder

# Utilities
import json

class JSON_Encoder(JSONEncoder):
    ''' Custom JSON serializer '''

    def default(self, obj):
        try:                # Attempt regular serialization
            obj = JSONEncoder.default(self, obj)
        except TypeError:   # Write as string on failure
            # TODO - [Logging] | Throw a warning when this occurs
            obj = str(obj)
            
            return obj


# Interface class
from .setting import Setting
from .redis_settings import Redis_Settings
# Utilities
import os, kombu, socket, time
from celery import Celery

class RabbitMQ_Settings(Setting):
    ''' Used to specify and validate RabbitMQ settings '''

    CONFIG_TYPE = 'rabbitmq_settings'

    # RabbitMQ Config
    USE_TASKS = os.environ.get('USE_TASKS', True)
    FORCE_START_RABBITMQ = os.environ.get('FORCE_START_RABBITMQ', True)
    FORCE_START_CELERY = os.environ.get('FORCE_START_CELERY', True)
    RABBITMQ_HOST = os.environ.get('RABBITMQ_HOST', 'localhost')
    RABBITMQ_PORT = os.environ.get('RABBITMQ_PORT', '5672')
    RABBITMQ_USERNAME = os.environ.get('RABBITMQ_USERNAME', 'guest')
    RABBITMQ_PASSWORD = os.environ.get('RABBITMQ_PASSWORD', 'guest')
    RABBITMQ_CONNECTION_STRING = os.environ.get('RABBITMQ_CONNECTION_STRING')
    RABBITMQ_INSTALLATION_PATH = os.environ.get('RABBITMQ_INSTALLATION_PATH', '/usr/local/opt/rabbitmq/sbin/rabbitmq-server')

    def __init__(self, use_tasks:bool=None, force_start_rabbitmq:bool=None, force_start_celery:bool=None, rabbitmq_host:str=None, 
                    rabbitmq_port:str=None, rabbitmq_username:str=None, rabbitmq_password:str=None, rabbitmq_connection_string:str=None,
                    rabbitmq_installation_path:str=None):

        if use_tasks: RabbitMQ_Settings.USE_TASKS = use_tasks
        if force_start_rabbitmq: RabbitMQ_Settings.FORCE_START_RABBITMQ = force_start_rabbitmq
        if force_start_celery: RabbitMQ_Settings.FORCE_START_CELERY = force_start_celery
        if rabbitmq_host: RabbitMQ_Settings.RABBITMQ_HOST = rabbitmq_host
        if rabbitmq_port: RabbitMQ_Settings.RABBITMQ_PORT = int(rabbitmq_port)
        if rabbitmq_username: RabbitMQ_Settings.RABBITMQ_USERNAME = rabbitmq_username
        if rabbitmq_password: RabbitMQ_Settings.RABBITMQ_PASSWORD = rabbitmq_password
        if rabbitmq_installation_path: RabbitMQ_Settings.RABBITMQ_INSTALLATION_PATH = rabbitmq_installation_path

        if rabbitmq_connection_string: 
            RabbitMQ_Settings.RABBITMQ_CONNECTION_STRING = rabbitmq_connection_string
        else:
            RabbitMQ_Settings.RABBITMQ_CONNECTION_STRING = f"amqp://{RabbitMQ_Settings.RABBITMQ_USERNAME}:{RabbitMQ_Settings.RABBITMQ_PASSWORD}@{RabbitMQ_Settings.RABBITMQ_HOST}:{RabbitMQ_Settings.RABBITMQ_PORT}/"

        # Allow redis to be force started from the application
        if RabbitMQ_Settings.USE_TASKS and RabbitMQ_Settings.FORCE_START_RABBITMQ and not RabbitMQ_Settings.check_rabbitmq_connection():
            print('Force starting RabbitMQ...')
            os.system(RabbitMQ_Settings.RABBITMQ_INSTALLATION_PATH + ' -detached'); time.sleep(4)
            print('Done :)')


    @staticmethod
    def get_log_data():
        ''' Returns a list of the settings to log to console '''

        rabbitmq_online = RabbitMQ_Settings.check_rabbitmq_connection()
        if RabbitMQ_Settings.USE_TASKS: rabbit_conn_test = 'Connected to RabbitMQ :)' if rabbitmq_online else 'WARNING - RabbitMQ ping failed. Ensure the service is running and config is correct. Defaulting to mutliprocess task engine'
        else: rabbit_conn_test = 'RabbitMQ is disabled via config, defaulting to multiprocess task engine'
       
        #celery_online = RabbitMQ_Settings.check_rabbitmq_connection()
        #if RabbitMQ_Settings.USE_TASKS: celery_conn_test = 'Connected to Celery :)' if celery_online else 'WARNING - Celery ping failed. Ensure the service is running and config is correct. Defaulting to mutliprocess task engine'
        #else: celery_conn_test = 'Celery is disabled via config, defaulting to multiprocess task engine'

        return [
            f'RabbitMQ connection string set to [{RabbitMQ_Settings.RABBITMQ_CONNECTION_STRING}]',
            rabbit_conn_test,
            #celery_conn_test
        ]

    @staticmethod
    def check_rabbitmq_connection() -> bool:
        ''' Check if RabbitMQ is online '''

        with kombu.Connection(RabbitMQ_Settings.RABBITMQ_CONNECTION_STRING) as c:
            try:
                c.connect()
            except socket.error:
                return False
            else:
                return True



# Redis
import redis

# Redis Settings
from ..config import Redis_Settings

# Redis Errors
from redis.exceptions import ResponseError, DataError

# Encoding
from ..encoder import JSON_Encoder
import json

# Utilities
import os

# Typing
from typing import Union

# TODO - [Stability]     | Dummy cache if Redis not running
# TODO - [Stability]     | Timeouts + Automatic retries
# TODO - [Useability]    | Allow serialization/deserialization and storage of typed data
# TODO - [Useability]    | Support for lists (queues)

class Cache:
    ''' Client for caching task results or database queries '''

    def __init__(self):
        # Connect to Redis
        self._redis = redis.Redis(host=Redis_Settings.REDIS_HOST, port=Redis_Settings.REDIS_PORT, db=Redis_Settings.REDIS_DB)


    def cache_string(self, key:str, value:str):
        ''' Add or update a key-value pair in the cache '''

        self._redis.set(key, value)


    def cache_dict(self, key:str, value:dict):
        ''' Add or overwrite a dictionary in the cache '''

        # Serialize dictionary then store
        self.cache_string(key, json.dumps(value, cls=JSON_Encoder)) 


    def cache_list(self, key:str, value:list):
        ''' Add or overwrite a list in the cache '''

        # Serialize the list with `json.dumps()` and store
        self.cache_dict(key, value)


    def cache_dynamic_dict(self, key:str, value:dict):
        ''' Store a python dictionary as a hash. Allows dictionary values to be updated without fetching the stored value '''

        # Attempt to insert the raw dictionary in Redis
        try:
            self._redis.hset(key, mapping=value)

        # Serialize the inner contents if necessary
        except DataError:
            value = {x: json.dumps(y, cls=JSON_Encoder) for x,y in value.items()}
            self._redis.hset(key, mapping=value)


    def get_dynamic_dict_value(self, dict_key:str, key:str) -> str:
        ''' Get a value from a cached dictionary stored with cache_dynamic_dict() '''

        # Retrieve the dynamically cached dictionary
        result = self._redis.hget(dict_key, key)
        
        if result:
            return result.decode()


    def clear_dynamic_dict_value(self, dict_key:str, key:Union[list,str]):
        ''' Delete a key or list of keys from a cached dictionary stored with cache_dynamic_dict() '''

        self._redis.hdel(dict_key, *([key] if isinstance(key,str) else key))


    def get(self, key:str) -> str:
        ''' Fetch data from the cache by key '''

        # Retrieve the cached value and fix type if necessary
        try:
            return self._redis.get(key).decode()

        # Attempt to retrieve a dynamically cached dictionary on failure
        except ResponseError: 
            return {x.decode(): y.decode() for x,y in  self._redis.hgetall(key).items()}


    def remove(self, key:Union[list,str]):
        ''' Remove a stored key or list of keys '''

        self._redis.delete(*([key] if isinstance(key,str) else key))


    def keys(self, regex:str='*') -> list:
        ''' View the keys stored in the cache or all keys matching a passed regular expression'''

        return [k.decode() for k in self._redis.keys(regex)]


    def view(self, regex:str='*') -> dict:
        ''' View the entire contents of the cache or the keys+contents matching a passed regular expression '''
        
        return {k:self.get(k) for k in self.keys(regex)}


    def flush(self, force=False):
        ''' Flush the cache '''

        self._redis.flushall(asynchronous=(not force))


    def save_to_disk(self):
        ''' Force-save the cache to disk '''

        self._redis.save()
        



# Individual configs
from . import App_Settings, MongoDB_Settings, RabbitMQ_Settings, Redis_Settings

# Utilities
import os

# Typing
from typing import Union

class Settings(App_Settings, MongoDB_Settings, RabbitMQ_Settings, Redis_Settings):
    ''' Used to specify and validate all application settings '''

    def __init__(self, 
                 app_settings:Union[dict, App_Settings]=None, 
                 mongodb_settings:Union[dict, MongoDB_Settings]=None, 
                 rabbitmq_settings:Union[dict, RabbitMQ_Settings]=None,
                 redis_settings:Union[dict, Redis_Settings]=None):


        if isinstance(app_settings, dict): App_Settings.from_dict('app_settings', app_settings)
        if app_settings == None: App_Settings()

        if isinstance(mongodb_settings, dict): MongoDB_Settings.from_dict('mongodb_settings', mongodb_settings)
        if mongodb_settings == None: MongoDB_Settings()
        
        if isinstance(rabbitmq_settings, dict): RabbitMQ_Settings.from_dict('rabbitmq_settings', rabbitmq_settings)
        if rabbitmq_settings == None: RabbitMQ_Settings()
        
        if isinstance(redis_settings, dict): Redis_Settings.from_dict('redis_settings', redis_settings)
        if redis_settings == None: Redis_Settings()


    @staticmethod
    def log_config():
        ''' Log 3rd party config options '''

        to_log = [App_Settings, MongoDB_Settings, RabbitMQ_Settings, Redis_Settings]
        for config in to_log: 
            for option in config.get_log_data(): print(' * [dead-simple-config]: ' + option)
            print('--')




"""Setup for Dead Simple Framework"""

import setuptools

with open('README.md') as f:
    README = f.read()

setuptools.setup(
    author="Peter Swanson",
    author_email="peter.swanson@ucdavis.edu",
    name='dead_simple_framework',
    license="MIT",
    description='Specify APIs with a Python dict and run with one line',
    version='v0.4.4',
    long_description=README,
    url='https://github.com/Topazoo/dead_simple_framework',
    packages=setuptools.find_packages(),
    python_requires=">=3.5",
    install_requires=['flask', 'pymongo', 'celery', 'flask-cors', 'requests', 'redis', 'eventlet', 'pyOpenSSL'],
    long_description_content_type='text/markdown',
    classifiers=[]
)


''' Dead Simple Framework - Specify APIs with a Python dict and run with one line

    Author: Peter Swanson
'''

from .main import Application
from .database import Database
from .tasks import Task_Manager
from .api.client import API
from .cache import Cache
from .config import Route as Route, TaskConfig as Task

from .sample import sample_config


# MongoDB
from pymongo import MongoClient
from pymongo.collection import Collection

# MongoDB Settings
from ..config import MongoDB_Settings

# Utilities
import os

# TODO - [Stability]     | Warnings + (Dummy `Database` if MongoDB isn't running)?
# TODO - [Stability]     | Timeouts + Automatic retries
# TODO - [Extendability] | Initial indexes, host and other config from main app config
# TODO - [Useability]    | Allow `connect()` and `disconnect()` as class methods

class Database:
    ''' Client to store and retrieve dictionary data from the database '''

    # Default database/collection pulled from the environment but defaults to these
    CONNECTION = None

    def __init__(self,database:str=None, collection:str=None):
        # For use with the context manager (`with` statement)
        self.database = database
        self.collection = collection


    def connect(self, database:str=None, collection:str=None) -> Collection:
        ''' Connect to a database and collection and return the collection

            - `database` defualts to the `MONGODB_DEFAULT_DB` setting if not passed

            - `collection` defualts to the `MONGODB_DEFAULT_COLLECTION` setting if not passed
        '''

        # Connect to the default database/collection if none passed
        if not database:   database   = MongoDB_Settings.MONGODB_DEFAULT_DB if not self.database else self.database
        if not collection: collection = MongoDB_Settings.MONGODB_DEFAULT_COLLECTION if not self.collection else self.collection

        # If a connection to the client isn't open, open one
        if not self.CONNECTION: self.CONNECTION = MongoClient(MongoDB_Settings.MONGODB_CONNECTION_STRING)

        # Return the collection
        return self.CONNECTION.get_database(database).get_collection(collection)
    

    def disconnect(self):
        ''' Disconnect from a database and collection '''

        if self.CONNECTION:
            self.CONNECTION.close()
            self.CONNECTION = None


    def __enter__(self) -> Collection:
       return self.connect()
       

    def __exit__(self, exception_type, exception_value, traceback):
        self.disconnect()



# Celery
from celery import Celery, Task
from celery.result import AsyncResult

# Celery Configuration
from celery.schedules import crontab
from kombu.serialization import register

# Task config class
from ..config import TaskConfig

# Task Settings
from ..config import Redis_Settings, RabbitMQ_Settings

# Custom Tasks
from .task import Database_Task, Store_Task, Store_Latest_Task

# Database
from ..database import Database
from bson import ObjectId

# Cache
from ..cache import Cache

# Utilities
from .utils import upsert_persistently_and_cache
from celery import chain, group
import os, json

# Typing
from typing import Union, Dict

# Debug
import logging

# TODO - [Stability]     | Warnings + (Dummy `Task_Manager` if RabbitMQ isn't running [use multiprocess])?
# TODO - [Stability]     | RPC backend if Redis isn't running
# TODO - [Stability]     | Timeouts for all tasks (especially sync tasks)
# TODO - [Useability]    | Retreival method for failed tasks
# TODO - [Useability]    | Use `Celery` via adapter to hide unwanted methods/attributes


class Task_Manager(Celery):
    ''' Client for managing asynchronous tasks 
    
        Allows any tasks specified in the application configuration dictionary's
        `tasks` section to be run or scheduled. Also allows retreival of the most
        recent result of a task. Handles periodic tasks too
    '''

    _app = None                                    # Internal application reference
    _internal_tasks:Dict[str, TaskConfig] = {}     # Internal reference to all dynamically registered tasks
   
    _results_collection = Database_Task._collection     # Collection for task results
    _results_cache_key  = Database_Task._cache_key      # Cache key for latest task result ID storage

    def __init__(self, dynamic_tasks:dict=None, main=None, loader=None, backend=None, amqp=None, events=None, log=None, control=None, set_as_current=True, tasks=None, broker=None, include=None, changes=None, config_source=None, fixups=None, task_cls=None, autofinalize=True, namespace=None, strict_typing=True, result_persistent = False, ignore_result=False, **kwargs):
        # Attach to RabbitMQ to allow tasks to be sent to/processed by any available worker
        broker, backend = RabbitMQ_Settings.RABBITMQ_CONNECTION_STRING, Redis_Settings.REDIS_CONNECTION_STRING
        
        # Override path to tasks to simplify invoking Celery from the command line
        super().__init__(main='dead_simple_framework', loader=loader, backend=backend, amqp=amqp, events=events, log=log, control=control, set_as_current=set_as_current, tasks=tasks, broker=broker, include=include, changes=changes, config_source=config_source, fixups=fixups, task_cls=task_cls, autofinalize=autofinalize, namespace=namespace, strict_typing=strict_typing, **kwargs)
        
        # Set Pickle as the task result serializer
        self.configure()

        # Register all tasks specified in the `tasks` section of the application config
        if dynamic_tasks:
            self.register_tasks(dynamic_tasks)

        Task_Manager._app = self

    def configure(self):
        ''' Optimize Celery configuration '''

        # Optimize the Celery configuration and set results serializer
        self.conf.update(
            task_create_missing_queues = True,
            task_acks_late = True,
            worker_prefetch_multiplier = 1,
            result_expires = 1,
            task_serializer='pickle',
            result_serializer='pickle',
            accept_content=['pickle']
        )

    def register_tasks(self, tasks:dict):
        ''' Dynamically register all tasks specified in the `tasks` section of the application config with Celery '''

        # Loop over tasks passed in the config
        for task_name, task_config in tasks.items():
            # Create a task config object from a dictionary if a dictionary based config was passed 
            task = self._cast_task(task_name, task_config)

            # Create and register a new Celery task with auto-caching results unless explicitely specified otherwise
            task.set_task(self.task(task.logic, name=task_name, result_serializer='pickle', base=self.get_task_type(task.depends_on), ignore_result=True if task.schedule else False))
            
            # Store a reference in the Task_Manager
            self._internal_tasks[task_name] = task

            # If the task should run on a schedule, set that up
            if task.schedule != None:
                self.register_periodic_task(task)
                

    def register_task_chain(self, task:TaskConfig):
        ''' Register a chain of tasks as a single task. Allows top-level tasks to be called that depend on the results of sub-tasks '''

        # Create a new task that invokes the chain of tasks when executed
        chain_as_task = TaskConfig(name=task.name + '_chain', logic=lambda x=None: self.chain(task.name, task.default_args or [], task.default_kwargs or {}))
        chain_as_task.set_task(self.task(chain_as_task.logic, name=chain_as_task.name, result_serializer='pickle', base=self.get_task_type(task.depends_on), ignore_result=False))
        
        # Store an internal reference to the task chain's top-level task
        self._internal_tasks[chain_as_task.name] = chain_as_task

        return chain_as_task.name


    def register_periodic_task(self, task:TaskConfig):
        ''' Dynamically register a task to run periodically via Celery '''

        # Create a schedule dictoinary in Celery if one does not exist
        if not self.conf.beat_schedule: self.conf.beat_schedule = {}

        # Determine if this scheduled task is a chain and register as such if it is
        task_name = self.register_task_chain(task) if task.depends_on else task.name
        
        # Add the task to the Celery schedule
        self.conf.beat_schedule[task.name] = {
            'task': task_name,
            'schedule': crontab(**task.schedule),
            'args': task.default_args
        }


    @classmethod
    def schedule_task(cls, task_name:str, *args, **kwargs):
        ''' Schedule an asynchronous task to be run by the next available worker.
            The latest result for a task can be retrieved with `get_result()` '''
        
        # Check to see if the relies on sub-tasks and must be chained 
        if cls._internal_tasks[task_name].depends_on:
            return cls._app.chain(task_name, args, kwargs)
        
        # Apply default arguments if none provided
        default_args = cls._internal_tasks[task_name].default_args
        if not args and default_args:
            args = (default_args,)

        # Drop the result in Celery/Redis as we're relying on the framework to cache them
        return cls._app.send_task(task_name, *args, **kwargs, ignore_result=True)


    @classmethod
    def run_task(cls, task_name:str, sync=True, *args, **kwargs):
        ''' Run an asynchronous task. Gets the result immediately if sync=True (force synchronous)'''

        if not sync: return cls.schedule_task(task_name, *args, **kwargs)

        cache = Cache()

        is_started = False
        last_id = cache.get_dynamic_dict_value(cls._results_cache_key, task_name)
        # Wait for ID of cached result to update then return
        while cache.get_dynamic_dict_value(cls._results_cache_key, task_name) == last_id:
            if not is_started:
                # Send the task to the next available worker once
                cls.schedule_task(task_name, *args, **kwargs)
                is_started = True

        # Fetch the result from the cache and return
        return cls.get_result(task_name)


    @classmethod
    def chain(cls, task_name:str, *args, **kwargs) ->list:
        ''' Form a chain of dependant sub-tasks for a given task '''

        # Get the top level task and its parameters
        task = cls._internal_tasks[task_name]

        # Begin a list of dependent tasks (executed first to last)
        dependants = [task.task.s(*(task.default_args or []), **(task.default_kwargs or {}))] 

        # Add each sub-task to the list of dependant tasks to be executed before the task after it
        while task.depends_on:
            task_name   = task.depends_on
            task        = cls._internal_tasks[task_name]
            dependants.insert(0, task.task.s(*(task.default_args or []), **(task.default_kwargs or {}))) 

        # Create a chain of tasks allowing the argument of the first to be passed to the next and so on
        return chain(*dependants)()


    @classmethod
    def parallelize(cls, tasks, cache_as:str=None):
        ''' Takes a list of tasks (in the same format as `schedule_task()`) and runs them in parallel 
            If `cache_as` is set, the result will be cached as if it were a task and will be accessible via `
        '''

        to_run = []
        for task in tasks: # Create signature (with args) for every task being run in parallel
            task_name, args, kwargs = task[0], task[1] if len(task) > 1 else [], task[2] if len(task) > 2 else {}
            task = cls._internal_tasks[task_name]
            to_run.append(task.task.s(*args, **kwargs))

        # Run the task and immediately get the result
        task = group(to_run).apply_async()

        # Allow sync subtasks in the event `parallelize` is used within another task
        result = task.get(disable_sync_subtasks=False)

        # Cache the data like a regular task if specified
        if cache_as:
            data = {'task_name': cache_as, 'task_id': str(task.id), 'task_result': result}
            upsert_persistently_and_cache(cls._results_collection, cls._results_cache_key, data, cache_as)
        
        return result


    @classmethod 
    def get_result(cls, task_name:str):
        ''' Get the latest result of a task if it exists
        
        Args:

            task_name (str): The name of the task to retrieve the last result for

        Returns:

           The last stored result for the task
        '''

        # Get the database record ID of the task result from the cache
        result_id = Cache().get_dynamic_dict_value(cls._results_cache_key, task_name)
        if result_id: # Use it to retrieve the result
            with Database(collection=cls._results_collection) as db:
                res =  db.find_one({'_id': ObjectId(result_id)})
                return res['task_result'] if res else res


    @classmethod 
    def get_results(cls, task_name:str):
        ''' Get all stored results for a task 
        
        Args:

            task_name (str): The name of the task to retrieve results for

        Returns:

            All stored results for the task. Only the latest result will be retuned if the task a `Store_Latest_Task` task
        '''

        with Database(collection=cls._results_collection) as db:
            return [res['task_result'] for res in list(db.find({'task_name': task_name}))]


    @staticmethod
    def _cast_task(name:str, task_config:Union[dict, TaskConfig]) -> TaskConfig:
        ''' Takes a task config dictionary and casts it to a `TaskConfig` object 
        
        Args:

            name (str): The name for this task

            task_config (Union[dict, TaskConfig]): The configuration for the task. Can be passed a config dictionary
                or an already created TaskConfig. If the latter if passed, no conversion is necessary
        
        Returns:

            A `Route` object representing the config for that route
        '''
        
        return task_config if isinstance(task_config, TaskConfig) else TaskConfig.from_dict(name, task_config)


    def get_task_type(self, should_store:bool) -> Task:
        ''' Get the type of task to use (different task types store results in different ways)

        Args:

            should_store (bool): Whether or not results should be stored. The type is based on the `store_results` key, if it is present, possible values:
                - None     | Store the last result in the database [default]
                - True     | Store all results in the database
                - False    | Store no results in the database

        Returns:

            The celery task type that should be used to instantiate a given task with based on it's `should_store` attribute
        '''

        if should_store: 
            return Store_Task
        if should_store == False:
            return Task

        return Store_Latest_Task


# Storage Utilities
from collections import defaultdict


class API_Error(Exception):
    ''' Error thrown internally by the server API when a bad request is received '''
    
    def __init__(self, message, code):
        super(Exception, self).__init__(message)
        self.code = code


class API_Client_Error(Exception):
    ''' Error thrown by the API client when an error is received '''

    # TODO - Full list of HTTP error codes
    CODES = defaultdict(lambda: 'Unknown', {
        404: 'Resource Not Found',
        403: 'Forbidden - Ensure proper credentials were supplied',
        405: 'Not supported - The method is not supported',
        500: 'Server Error - The endpoint encountered an error processing the request'
    })
    
    
    def __init__(self, url:str, method:str, code:int):
        super(Exception, self).__init__(f"{method} Request Error [{code}] | {self.CODES[code]} | {url}")
        self.code = code
        self.method = method



# Base class
from ..config import Config

class Setting(Config):
    ''' Interface shared by all settings classes '''

    CONFIG_TYPE = 'setting'

    @Config.validate_dict
    def from_dict(cls, _, config:dict) -> 'Setting':
        ''' Factory for generating a `Setting` class from a dictionary. Overriden by concrete settings classes '''
        
        return cls(**config)

    @staticmethod
    def get_log_data():
        ''' Returns a list of the settings to log to console 
            Overridden by concrete classes
        '''

        return []


sample_config = {
    '/demo': {
        'name': 'demo',
        'methods': ['GET', 'POST', 'DELETE', 'PUT'],
        'template': None,
        'defaults': None,
        'logic': None,
        'collection': 'demo'
    }
}
